///////////////////////////////////////////////////////////////////////////////
// File: ServerMain.cpp
///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007 JetByte Limited.
//
// This software is provided "as is" without a warranty of any kind. All
// express or implied conditions, representations and warranties, including
// any implied warranty of merchantability, fitness for a particular purpose
// or non-infringement, are hereby excluded. JetByte Limited and its licensors
// shall not be liable for any damages suffered by licensee as a result of
// using the software. In no event will JetByte Limited be liable for any
// lost revenue, profit or data, or for direct, indirect, special,
// consequential, incidental or punitive damages, however caused and regardless
// of the theory of liability, arising out of the use of or inability to use
// software, even if JetByte Limited has been advised of the possibility of
// such damages.
//
///////////////////////////////////////////////////////////////////////////////

#include "JetByteTools/Admin/Admin.h"
#include "JetByteTools/Admin/WinsockWrapper.h"

#include "SocketServer.h"

#include "JetByteTools/CoreTools/Exception.h"
#include "JetByteTools/CoreTools/FileUtils.h"
#include "JetByteTools/CoreTools/StringUtils.h"
#include "JetByteTools/CoreTools/DebugTrace.h"
#include "JetByteTools/CoreTools/StringConverter.h"
#include "JetByteTools/CoreTools/GlobalErrorHandler.h"

#include "JetByteTools/IOTools/AsyncFileLog.h"
#include "JetByteTools/SocketTools/FullAddress.h"
#include "JetByteTools/SocketTools/AddressTypeIPv4.h"
#include "JetByteTools/SocketTools/DatagramServerSocketAllocator.h"

#include "ServerCommon/IOPool.h"
#include "ServerCommon/BufferAllocator.h"
#include "ServerCommon/CommandLine.h"
#include "ServerCommon/SimpleServerShutdownHandler.h"

#pragma hdrstop

#include <iostream>

///////////////////////////////////////////////////////////////////////////////
// Using directives
///////////////////////////////////////////////////////////////////////////////

using JetByteTools::Core::_tstring;
using JetByteTools::Core::CException;
using JetByteTools::Core::OutputEx;
using JetByteTools::Core::CDebugTrace;
using JetByteTools::Core::CStringConverter;
using JetByteTools::Core::CGlobalErrorHandler;

using JetByteTools::Socket::CDatagramServerSocketAllocator;
using JetByteTools::Socket::CFullAddress;
using JetByteTools::Socket::CAddressTypeIPv4;
using JetByteTools::Socket::CAddressType;
using JetByteTools::Socket::CConnectionLimiter;
using JetByteTools::Socket::ListenBacklog;

using JetByteTools::IO::CAsyncFileLog;

using std::cerr;
using std::endl;

///////////////////////////////////////////////////////////////////////////////
// Program entry point
///////////////////////////////////////////////////////////////////////////////

int main(int argc, const char **argv)
{
   CGlobalErrorHandler errorHandler;

   int result = 0;

   try
   {
      CAsyncFileLog log(_T("EchoServerUDPMultiCast.log"));

      CDebugTrace::LogInstaller logInstaller(log);

      try
      {
         CCommandLine commandLine(
            argc,
            argv,
            _T("EchoServerUDPMultiCast"),
            CCommandLine::UDPMulticastServer);

         if (commandLine.Parse())
         {
            CIOPool pool(
               commandLine.NumberOfIOThreads(),
               commandLine.DisplayDebug());

            pool.Start();

            CDatagramServerSocketAllocator socketAllocator(
               commandLine.SocketPoolSize(),
               commandLine.MaxBuffersPooledPerSocket(),
               commandLine.MaxConsecutiveOperationsOnOneSocket());

            CBufferAllocator bufferAllocator(
               commandLine.BufferSize(),
               commandLine.BufferPoolSize(),
               commandLine.DisplayDebug());

            const CAddressType addressPreference = CAddressTypeIPv4::instance;

            const CFullAddress address(
               commandLine.Server(),
               commandLine.Port(),
               addressPreference);

            const ListenBacklog listenBacklog = commandLine.ListenBacklog();

            CConnectionLimiter connectionLimiter(commandLine.MaxConnections());

            CSocketServer server(
               address,
               listenBacklog,
               pool,
               socketAllocator,
               bufferAllocator,
               connectionLimiter);

            server.Start();

            server.StartAcceptingConnections();

            if (commandLine.HasMulticastAddress())
            {
               const CFullAddress groupAddress(
                  commandLine.MulticastAddress(),
                  commandLine.Port());

               const CFullAddress interfaceAddress(
                  commandLine.Server(),
                  0);

               server.JoinMulticastGroup(          // repeat for all multicast groups that you wish to join
                  groupAddress,
                  interfaceAddress);
            }

            CSimpleServerShutdownHandler shutdownHandler(server, commandLine.EventNamePrefix());

            shutdownHandler.WaitForShutdownRequest();

            server.WaitForShutdownToComplete();

            pool.WaitForShutdownToComplete();

            socketAllocator.ReleaseSockets();

            bufferAllocator.Flush();
         }
         else
         {
            result = 1;
         }
      }
      catch(const CException &e)
      {
         OutputEx(_T("Exception: ") + e.GetDetails());

         result = 1;
      }
      catch(...)
      {
         OutputEx(_T("Unexpected exception"));

         result = 1;
      }
   }
   catch(const CException &e)
   {
      cerr << "Exception: " << e.GetDetailsA() << endl;

      result = 1;
   }
   catch(...)
   {
      cerr << "Unexpected exception" << endl;

      result = 1;
   }

   return result;
}

///////////////////////////////////////////////////////////////////////////////
// End of file: ServerMain.cpp
///////////////////////////////////////////////////////////////////////////////
