///////////////////////////////////////////////////////////////////////////////
// File: TestMain.cpp
///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007 JetByte Limited.
//
// This software is provided "as is" without a warranty of any kind. All
// express or implied conditions, representations and warranties, including
// any implied warranty of merchantability, fitness for a particular purpose
// or non-infringement, are hereby excluded. JetByte Limited and its licensors
// shall not be liable for any damages suffered by licensee as a result of
// using the software. In no event will JetByte Limited be liable for any
// lost revenue, profit or data, or for direct, indirect, special,
// consequential, incidental or punitive damages, however caused and regardless
// of the theory of liability, arising out of the use of or inability to use
// software, even if JetByte Limited has been advised of the possibility of
// such damages.
//
///////////////////////////////////////////////////////////////////////////////

#include "JetByteTools/Admin/Admin.h"
#include "JetByteTools/Admin/WinsockWrapper.h"

#include "AsyncSocketConnectionManager.h"

#include "JetByteTools/CoreTools/Exception.h"
#include "JetByteTools/CoreTools/FileUtils.h"
#include "JetByteTools/CoreTools/StringUtils.h"
#include "JetByteTools/CoreTools/DebugTrace.h"
#include "JetByteTools/CoreTools/StringConverter.h"
#include "JetByteTools/Win32Tools/GlobalName.h"
#include "JetByteTools/Win32Tools/ManualResetEvent.h"

#include "JetByteTools/IOTools/IOPool.h"
#include "JetByteTools/IOTools/BufferAllocator.h"
#include "JetByteTools/IOTools/SmartBuffer.h"

#include "JetByteTools/SocketTools/DatagramSocketAllocator.h"
#include "JetByteTools/SocketTools/AddressIPv4.h"

#pragma hdrstop

#include <iostream>

///////////////////////////////////////////////////////////////////////////////
// Using directives
///////////////////////////////////////////////////////////////////////////////

using JetByteTools::Core::CException;
using JetByteTools::Core::Output;
using JetByteTools::Core::OutputEx;
using JetByteTools::Core::ToString;
using JetByteTools::Core::_tstring;
using JetByteTools::Core::CStringConverter;

using JetByteTools::Win32::CGlobalName;
using JetByteTools::Win32::CManualResetEvent;

using JetByteTools::IO::CIOPool;
using JetByteTools::IO::CBufferAllocator;
using JetByteTools::IO::IBuffer;
using JetByteTools::IO::CSmartBuffer;

using JetByteTools::Socket::CAddressIPv4;
using JetByteTools::Socket::CSmartDatagramSocket;
using JetByteTools::Socket::CDatagramSocketAllocator;

using std::cout;
using std::cerr;
using std::endl;

///////////////////////////////////////////////////////////////////////////////
// Program entry point
///////////////////////////////////////////////////////////////////////////////

int main(int argc, char *argv[ ])
{
   int result = 1;

   if (argc < 4 || argc > 7)
   {
      cerr << "usage: EchoServerUDPMultiCastTest.exe <multicast dotted.ip.address> <port> <interface dotted.ip.address> [numberOfConnections (1000)] [number of 1024 byte messages (1000)] [pause 1 sec after x connect attempts (1000)]" << endl;
      exit(1);
   }

   const char *pMulticastDottedIp = argv[1];

   const unsigned short multicastPort = static_cast<unsigned short>(atol(argv[2]));

   const char *pInterfaceDottedIp = argv[3];

   size_t numConnections = 1000;

   if (argc >= 5)
   {
      numConnections = atol(argv[4]);
   }

   size_t numMessages = 1000;

   if (argc >= 6)
   {
      numMessages = atol(argv[5]);
   }

   size_t delayAfter = 1000;

   if (argc >= 7)
   {
      delayAfter = atol(argv[6]);
   }


   try
   {
      const size_t maxSocketsInPool = 10;
      const size_t maxPooledBuffersPerSocket = 10;
      const size_t maxConsecutiveOperations = 10;
      const size_t maxBuffersInPool = 10;
      const size_t bufferSize = 5000;

      CDatagramSocketAllocator socketAllocator(
         maxSocketsInPool,
         maxPooledBuffersPerSocket,
         maxConsecutiveOperations);

      CBufferAllocator bufferAllocator(
         bufferSize,
         maxBuffersInPool);

      CIOPool ioPool(
         0);                           // number of threads (0 = 2 x processors)

      ioPool.Start();

      CAsyncSocketConnectionManager connector(
         ioPool,
         socketAllocator,
         bufferAllocator);

      const CAddressIPv4 address(CStringConverter::AtoT(pMulticastDottedIp), multicastPort);

      const CAddressIPv4 interfaceAddress(CStringConverter::AtoT(pInterfaceDottedIp), 0);

      const int ttl = 10;

      try
      {
         cout << "Creating " << numConnections << " connections" << endl;

         for (size_t i = 0; i < numConnections; ++i)
         {
            CSmartBuffer message(bufferAllocator.Allocate());

            message->Use(50);

            CSmartDatagramSocket socket = connector.Bind(CAddressIPv4());

            socket->SetMulticastSendInterface(interfaceAddress, ttl);

            socket->SendTo(address, message);
            socket->Read();

            if (i && i % delayAfter == 0)
            {
               OutputEx(ToString(i) + _T(" connections active"));

               Sleep(1000);
            }
         }

         OutputEx(_T("All connections active"));

         CManualResetEvent shutdownEvent(CGlobalName(_T("JetByteToolsServerShutdown")));

         HANDLE handlesToWaitFor[2];

         handlesToWaitFor[0] = shutdownEvent.GetWaitHandle();
//         handlesToWaitFor[1] = messages.GetWaitHandle();

         DWORD waitResult = ::WaitForMultipleObjects(1, handlesToWaitFor, false, INFINITE);

         if (waitResult == WAIT_OBJECT_0)
         {
            result = 1;
         }
         else if (waitResult == WAIT_OBJECT_0 + 1)
         {
            result = 1;//messages.CheckMessages() ? 0 : 1;
         }
         else
         {
            throw CException(_T("main()"), _T("Unexpected result from WaitForMultipleObjects"));
         }

         cout << "All connections complete" << endl;
      }
      catch(const CException &e)
      {
         cout << "Exception: " << e.GetDetailsA() << endl;

         result = 1;
      }
      catch(...)
      {
         cout << "Unexpected exception" << endl;

         result = 1;
      }

      ioPool.WaitForShutdownToComplete();

      socketAllocator.ReleaseSockets();

      bufferAllocator.Flush();
   }
   catch(const CException &e)
   {
      cout << "Exception: " << e.GetDetailsA() << endl;

      result = 1;
   }
   catch(...)
   {
      cout << "Unexpected exception" << endl;

      result = 1;
   }

   cout << "Test " << (result == 0 ? "Passed" : "Failed") << endl;

   return result;
}

///////////////////////////////////////////////////////////////////////////////
// End of file: ServerMain.cpp
///////////////////////////////////////////////////////////////////////////////
