///////////////////////////////////////////////////////////////////////////////
// File: CommandLine.cpp
///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2006 JetByte Limited.
//
// This software is provided "as is" without a warranty of any kind. All
// express or implied conditions, representations and warranties, including
// any implied warranty of merchantability, fitness for a particular purpose
// or non-infringement, are hereby excluded. JetByte Limited and its licensors
// shall not be liable for any damages suffered by licensee as a result of
// using the software. In no event will JetByte Limited be liable for any
// lost revenue, profit or data, or for direct, indirect, special,
// consequential, incidental or punitive damages, however caused and regardless
// of the theory of liability, arising out of the use of or inability to use
// software, even if JetByte Limited has been advised of the possibility of
// such damages.
//
///////////////////////////////////////////////////////////////////////////////

#include "JetByteTools/Admin/Admin.h"

#include "CommandLine.h"

#include "JetByteTools/CoreTools/Exception.h"
#include "JetByteTools/CoreTools/StringConverter.h"
#include "JetByteTools/CoreTools/FileUtils.h"
#include "JetByteTools/CoreTools/StringUtils.h"
#include "JetByteTools/CoreTools/Tchar.h"

#pragma hdrstop

#include <iostream>

///////////////////////////////////////////////////////////////////////////////
// Using directives
///////////////////////////////////////////////////////////////////////////////

using JetByteTools::Milliseconds;

using JetByteTools::Core::_tstring;
using JetByteTools::Core::CException;
using JetByteTools::Core::CStringConverter;
using JetByteTools::Core::ToString;

using JetByteTools::Socket::ConsecutiveSocketOperations;
using JetByteTools::Socket::PerSocketBufferPoolingLimit;

using std::cout;
using std::endl;

///////////////////////////////////////////////////////////////////////////////
// Constants
///////////////////////////////////////////////////////////////////////////////

static const _tstring s_serverArgument                   = _T("server");
static const _tstring s_portArgument                     = _T("port");
static const _tstring s_repeatArgument                   = _T("repeat");
static const _tstring s_packetsArgument                  = _T("packets");
static const _tstring s_packetSizeArgument               = _T("packetSize");
static const _tstring s_packetBatchSizeArgument          = _T("packetBatchSize");
static const _tstring s_packetBatchDelayArgument         = _T("packetBatchDelay");
static const _tstring s_preallocateSwitch                = _T("preallocate");
static const _tstring s_spinCountArgument                = _T("spinCount");
static const _tstring s_readTimeoutArgument              = _T("readTimeout");
static const _tstring s_maxBuffersPooledPerSocket        = _T("perSocketBufferPoolSize");
static const _tstring s_maxConsecutiveOperations         = _T("maxOps");
static const _tstring s_displayRTTSwitch                 = _T("displayRTT");
static const _tstring s_dataFlowTimeoutArgument          = _T("dataFlowTimeout");
static const _tstring s_displayStatusEveryArgument       = _T("displayStatusEvery");

///////////////////////////////////////////////////////////////////////////////
// CCommandLine
///////////////////////////////////////////////////////////////////////////////

CCommandLine::CCommandLine()
{
   m_parser.AddArgument(s_serverArgument);
   m_parser.AddArgument(s_portArgument);
   m_parser.AddArgument(s_repeatArgument);
   m_parser.AddArgument(s_packetsArgument);
   m_parser.AddArgument(s_packetSizeArgument);
   m_parser.AddArgument(s_packetBatchSizeArgument);
   m_parser.AddArgument(s_packetBatchDelayArgument);
   m_parser.AddArgument(s_spinCountArgument);
   m_parser.AddArgument(s_readTimeoutArgument);
   m_parser.AddArgument(s_dataFlowTimeoutArgument);
   m_parser.AddArgument(s_displayStatusEveryArgument);
   m_parser.AddArgument(s_maxBuffersPooledPerSocket);
   m_parser.AddArgument(s_maxConsecutiveOperations);

   m_parser.AddSwitch(s_preallocateSwitch);
   m_parser.AddSwitch(s_displayRTTSwitch);
}

bool CCommandLine::Parse(
   const DWORD argc,
   const char **argv)
{
   _tstring message;

   bool ok = m_parser.Parse(argc, argv, message);

   if (ok)
   {
      ok = m_parser.IsSet(s_serverArgument) && m_parser.IsSet(s_portArgument);
   }

   if (ok)
   {
      if (Repeats() == 0)
      {
         message = _T("It doesn't make sense to have `-repeats 0`; Repeating 0 times means sending no data");

         ok = false;
      }
   }

   if (ok)
   {
      if (m_parser.IsSet(s_maxBuffersPooledPerSocket))
      {
         const unsigned int value = _ttoi(m_parser.GetValueAsString(s_maxBuffersPooledPerSocket).c_str());

         if (value > std::numeric_limits<PerSocketBufferPoolingLimit>::max())
         {
            message = _T("Error: -perSocketBufferPoolSize value \"") + ToString(value) + _T("\" out of range, maximum is 255");

            ok = false;
         }
      }
   }

   if (ok)
   {
      if (m_parser.IsSet(s_maxConsecutiveOperations))
      {
         const unsigned int value = _ttoi(m_parser.GetValueAsString(s_maxConsecutiveOperations).c_str());
         
         if (value > std::numeric_limits<ConsecutiveSocketOperations>::max())
         {
            message = _T("Error: -maxOps value \"") + ToString(value) + _T("\" out of range, maximum is 255");

            ok = false;
         }
      }
   }

   if (!ok)
   {
      ShowHelp(message);
   }

   return ok ;
}

_tstring CCommandLine::Server() const
{
   return m_parser.GetValueAsString(s_serverArgument);
}

unsigned short CCommandLine::Port() const
{
   return static_cast<unsigned short>(_ttol(m_parser.GetValueAsString(s_portArgument).c_str()));
}

CCommandLine::Count CCommandLine::Repeats() const
{
   return StringToCount(m_parser.GetOptionalValueAsString(s_repeatArgument, _T("1")).c_str());
}

CCommandLine::Count CCommandLine::Packets() const
{
   return StringToCount(m_parser.GetOptionalValueAsString(s_packetsArgument, _T("100")).c_str());
}

CCommandLine::Count CCommandLine::PacketSize() const
{
   return StringToCount(m_parser.GetOptionalValueAsString(s_packetSizeArgument, _T("1024")).c_str());
}

CCommandLine::Count CCommandLine::PacketBatchSize() const
{
   return StringToCount(m_parser.GetOptionalValueAsString(s_packetBatchSizeArgument, _T("0")).c_str());
}

CCommandLine::Count CCommandLine::PacketBatchPause() const
{
   return StringToCount(m_parser.GetOptionalValueAsString(s_packetBatchDelayArgument, _T("0")).c_str());
}

bool CCommandLine::PreAllocate() const
{
   return m_parser.IsSet(s_preallocateSwitch);
}

CCommandLine::Count CCommandLine::SpinCount() const
{
   return StringToCount(m_parser.GetOptionalValueAsString(s_spinCountArgument, _T("0")).c_str());
}

Milliseconds CCommandLine::ReadTimeout() const
{
   return _ttol(m_parser.GetOptionalValueAsString(s_readTimeoutArgument, _T("0")).c_str());
}

bool CCommandLine::DumpRTT() const
{
   return m_parser.IsSet(s_displayRTTSwitch);
}

Milliseconds CCommandLine::DataFlowTimeout() const
{
   if (!m_parser.IsSet(s_dataFlowTimeoutArgument))
   {
      return INFINITE;
   }

   return _ttol(m_parser.GetOptionalValueAsString(s_dataFlowTimeoutArgument, _T("0")).c_str());
}

Milliseconds CCommandLine::DisplayStatusEvery() const
{
   if (!m_parser.IsSet(s_displayStatusEveryArgument))
   {
      return INFINITE;
   }

   return _ttol(m_parser.GetOptionalValueAsString(s_displayStatusEveryArgument, _T("0")).c_str());
}

PerSocketBufferPoolingLimit CCommandLine::MaxBuffersPooledPerSocket() const
{
   return static_cast<PerSocketBufferPoolingLimit>(_ttoi(m_parser.GetOptionalValueAsString(s_maxBuffersPooledPerSocket, _T("10")).c_str()));
}

ConsecutiveSocketOperations CCommandLine::MaxConsecutiveOperationsOnOneSocket() const
{
   return static_cast<ConsecutiveSocketOperations>(_ttoi(m_parser.GetOptionalValueAsString(s_maxConsecutiveOperations, _T("10")).c_str()));
}

void CCommandLine::ShowHelp(
   const _tstring &message)
{
   cout << endl;
   cout << "EchoServerUDPTest: v7.0 - Copyright (c) 2006 JetByte Limited" << endl;

   if (message.length() != 0)
   {
      cout << endl;
      cout << "Error: " << CStringConverter::TtoA(message) << endl;
      cout << endl;
      cout << "In command line: \"" << CStringConverter::TtoA(m_parser.GetParsedCommandLine()) << "\"" << endl;
      cout << endl;
   }

   cout << "Usage: EchoServerUDPTest -server xxx.xxx.xxx.xxx -port xxxx" << endl;
   cout << endl;
   cout << "Command line parameters:" << endl;
   cout << " r -server                  The server address to connect to (dotted ip or" << endl;
   cout << "                            IPV6 format)." << endl;
   cout << " r -port                    The port to connect to." << endl;
   cout << " o -packets                 The number of packets to send." << endl;
   cout << "                            Defaults to 100" << endl;
   cout << " o -packetSize              Size of each packet." << endl;
   cout << "                            Defaults to 1024" << endl;
   cout << " o -packetBatchSize         Batch packets in groups of X size." << endl;
   cout << "                            Defaults to 0 (no batching)" << endl;
   cout << " o -packetBatchDelay        Delay for Y milliseconds between each batch." << endl;
   cout << "                            Defaults to 0 (no delay)" << endl;
   cout << " o -repeat                  The number of times to send each sequence of packets." << endl;
   cout << "                            Defaults to 1, no repeats, send each sequence only once." << endl;
   cout << "                            Note that you can run a test with more packets echoed" << endl;
   cout << "                            per connection using less memory by setting high value" << endl;
   cout << "                            for -repeat and a low value for -packets." << endl;
   cout << " o -readTimeout             Fail the test if it takes longer than Y milliseconds" << endl;
   cout << "                            for an echoed packet to arrive back." << endl;
   cout << "                            Defaults to 0; no timeout" << endl;
   cout << " o -preallocate             Preallocate sockets and buffers before connecting." << endl;
   cout << " o -displayRTT              Display the round trip times for each packet." << endl;
   cout << " o -dataFlowTimeout         Fail the test if it takes longer than Y milliseconds" << endl;
   cout << "                            for all dataflow to complete." << endl;
   cout << "                            Defaults to no timeout" << endl;
   cout << " o -displayStatusEvery      During data flow, display the status of each connection" << endl;
   cout << "                            every Y milliseconds." << endl;
   cout << "                            Defaults to not displaying status." << endl;
   cout << " o -displayRTT              Display the round trip times for each packet." << endl;
   cout << " o -perSocketBufferPoolSize The number of I/O buffers pooled in the socket." << endl;
   cout << "                            Defaults to 10." << endl;
   cout << " o -maxOps                  The maximum number of consecutive operations" << endl;
   cout << "                            per socket. Defaults to 10." << endl;
   cout << endl;
   cout << " r = required, o = optional" << endl;
   cout << endl;
   cout << "See http://www.serverframework.com for more details." << endl;

}

CCommandLine::Count CCommandLine::StringToCount(
   const _tstring &stringValue)
{
   const size_t value = _ttol(stringValue.c_str());

   if (value > static_cast<size_t>(std::numeric_limits<CCommandLine::Count>::max()))
   {
      throw CException(_T("CCommandLine::StringToCount()"), _T("Value too big"));
   }

   return static_cast<Count>(value);
}

///////////////////////////////////////////////////////////////////////////////
// End of file: CommandLine.cpp
///////////////////////////////////////////////////////////////////////////////
