///////////////////////////////////////////////////////////////////////////////
// File: DatagramSocketConnectionManager.cpp
///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2006 JetByte Limited.
//
// This software is provided "as is" without a warranty of any kind. All
// express or implied conditions, representations and warranties, including
// any implied warranty of merchantability, fitness for a particular purpose
// or non-infringement, are hereby excluded. JetByte Limited and its licensors
// shall not be liable for any damages suffered by licensee as a result of
// using the software. In no event will JetByte Limited be liable for any
// lost revenue, profit or data, or for direct, indirect, special,
// consequential, incidental or punitive damages, however caused and regardless
// of the theory of liability, arising out of the use of or inability to use
// software, even if JetByte Limited has been advised of the possibility of
// such damages.
//
///////////////////////////////////////////////////////////////////////////////

#include "JetByteTools/Admin/Admin.h"

#include "DatagramSocketConnectionManager.h"
#include "TestDatagrams.h"
#include "ITestDatagram.h"

#include "JetByteTools/CoreTools/Exception.h"
#include "JetByteTools/CoreTools/FileUtils.h"
#include "JetByteTools/CoreTools/StringUtils.h"
#include "JetByteTools/CoreTools/StringConverter.h"
#include "JetByteTools/CoreTools/DebugTrace.h"
#include "JetByteTools/CoreTools/ErrorCodeToErrorMessage.h"

#include "JetByteTools/SocketTools/IDatagramSocket.h"
#include "JetByteTools/SocketTools/IAllocateDatagramSockets.h"
#include "JetByteTools/SocketTools/AddressRenderer.h"

#pragma hdrstop

#include <iostream>

///////////////////////////////////////////////////////////////////////////////
// Using directives
///////////////////////////////////////////////////////////////////////////////

using JetByteTools::Milliseconds;

using JetByteTools::Core::Output;
using JetByteTools::Core::OutputEx;
using JetByteTools::Core::DumpData;
using JetByteTools::Core::CException;
using JetByteTools::Core::_tstring;
using JetByteTools::Core::CStringConverter;
using JetByteTools::Core::ToStringA;
using JetByteTools::Core::ErrorCodeToErrorMessage;
using JetByteTools::Core::ILockableObject;

using JetByteTools::IO::IIOPool;
using JetByteTools::IO::CSmartBuffer;
using JetByteTools::IO::IAllocateBuffers;

using JetByteTools::Socket::IDatagramSocket;
using JetByteTools::Socket::IAllocateDatagramSockets;
using JetByteTools::Socket::IAddress;
using JetByteTools::Socket::ShutdownSend;
using JetByteTools::Socket::ShutdownReceive;
using JetByteTools::Socket::CAddressRenderer;
using JetByteTools::Socket::ConnectionClosureReason;
using JetByteTools::Socket::CFilteringDatagramSocketConnectionManager;
using JetByteTools::Socket::CReadTimeoutDatagramSocketConnectionFilter;

using std::cout;
using std::endl;

///////////////////////////////////////////////////////////////////////////////
// CDatagramSocketConnectionManager
///////////////////////////////////////////////////////////////////////////////

CDatagramSocketConnectionManager::CDatagramSocketConnectionManager(
   IIOPool &pool,
   IAllocateDatagramSockets &socketAllocator,
   IAllocateBuffers &bufferAllocator,
   CTestDatagrams &datagrams,
   const Milliseconds readTimeout)
   :  CFilteringDatagramSocketConnectionManager(*this, pool, socketAllocator, bufferAllocator),
      m_readTimeoutFilter(*this, m_timerQueue),
      m_readTimeout(readTimeout),
      m_datagrams(datagrams)
{

}

CDatagramSocketConnectionManager::~CDatagramSocketConnectionManager()
{
   WaitForShutdownToComplete();
}

void CDatagramSocketConnectionManager::WaitForShutdownToComplete()
{
   WaitForShutdownToComplete(INFINITE);
}

bool CDatagramSocketConnectionManager::WaitForShutdownToComplete(
   const Milliseconds timeout)
{
   const auto shutdownComplete = CFilteringDatagramSocketConnectionManager::WaitForShutdownToComplete(timeout);

   if (shutdownComplete)
   {
      m_timerQueue.WaitForShutdownToComplete();
   }

   return shutdownComplete;
}

void CDatagramSocketConnectionManager::OnRecvFromCompleted(
   IDatagramSocket &socket,
   const IAddress &address,
   CSmartBuffer &buffer)
{
   (void)socket;
   (void)address;

   //DEBUG_ONLY(Output(_T(" Local Address: ") + CAddressRenderer::AsString(socket.GetLocalAddress(), true)));
   //DEBUG_ONLY(Output(_T("Remote Address: ") + CAddressRenderer::AsString(address, true)));

   //DEBUG_ONLY(Output(DumpData(buffer->GetMemory(), buffer->GetUsed(), 60)));

   m_datagrams.ValidateReceivedData(buffer->GetMemory(), buffer->GetUsed());
}

void CDatagramSocketConnectionManager::OnSendToCompleted(
   IDatagramSocket &socket,
   const IAddress & /*address*/,
   CSmartBuffer & /*buffer*/)
{
   if (m_readTimeout != 0)
   {
      m_readTimeoutFilter.SetReadTimeout(socket, m_readTimeout, *this);
   }

   socket.RecvFrom();
}

void CDatagramSocketConnectionManager::OnSendToCompletionError(
   IDatagramSocket & /*socket*/,
   const IAddress & /*address*/,
   CSmartBuffer & /*buffer*/,
   const DWORD lastError)
{
   OnError(_T("OnSendToCompletionError"), lastError);

   m_datagrams.Fail();
}

void CDatagramSocketConnectionManager::OnError(
   const _tstring &message)
{
   ILockableObject::Owner lock(m_lock);

   m_errors[message]++;
}

void CDatagramSocketConnectionManager::OnError(
   const _tstring &message,
   const DWORD lastError)
{
   OnError(message + _T(": ") + ErrorCodeToErrorMessage(lastError));
}

void CDatagramSocketConnectionManager::OnConnectionClosure(
   IDatagramSocket & /*socket*/,
   const ConnectionClosureReason /*reason*/)
{
   // Suppress the debug message that the server common version gives us.
}

void CDatagramSocketConnectionManager::OnConnectionClosed(
   IDatagramSocket & /*socket*/)
{
   // Suppress the debug message that the server common version gives us.
}

void CDatagramSocketConnectionManager::OnTimer(
   IDatagramSocket & /*socket*/,
   CReadTimeoutDatagramSocketConnectionFilter::UserData /*userData*/)
{
   OnError(_T("Read timeout"));

   m_datagrams.Fail();
}

bool CDatagramSocketConnectionManager::DumpErrors() const
{
   auto ok = true;

   if (m_errors.size() != 0)
   {
      ok = false;

      cout << "Errors" << endl;

      for (auto it = m_errors.begin(); it != m_errors.end(); ++it)
      {
         cout << ToStringA(it->second) << " - " << CStringConverter::TtoA(it->first) << endl;
      }
   }

   return ok;
}

///////////////////////////////////////////////////////////////////////////////
// End of file: DatagramSocketConnectionManager.cpp
///////////////////////////////////////////////////////////////////////////////
