///////////////////////////////////////////////////////////////////////////////
// File: TestDatagram.cpp
///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2009 JetByte Limited.
//
// This software is provided "as is" without a warranty of any kind. All
// express or implied conditions, representations and warranties, including
// any implied warranty of merchantability, fitness for a particular purpose
// or non-infringement, are hereby excluded. JetByte Limited and its licensors
// shall not be liable for any damages suffered by licensee as a result of
// using the software. In no event will JetByte Limited be liable for any
// lost revenue, profit or data, or for direct, indirect, special,
// consequential, incidental or punitive damages, however caused and regardless
// of the theory of liability, arising out of the use of or inability to use
// software, even if JetByte Limited has been advised of the possibility of
// such damages.
//
///////////////////////////////////////////////////////////////////////////////

#include "JetByteTools/Admin/Admin.h"

#include "TestDatagram.h"

#include "JetByteTools/CoreTools/Exception.h"
#include "JetByteTools/CoreTools/FileUtils.h"
#include "JetByteTools/CoreTools/StringUtils.h"
#include "JetByteTools/CoreTools/DebugTrace.h"

#pragma hdrstop

#include <algorithm>    // for min

///////////////////////////////////////////////////////////////////////////////
// Using directives
///////////////////////////////////////////////////////////////////////////////

using JetByteTools::Core::_tstring;
using JetByteTools::Core::CException;
using JetByteTools::Core::Output;
using JetByteTools::Core::DumpData;
using JetByteTools::Core::ToString;

using std::min;

///////////////////////////////////////////////////////////////////////////////
// CTestDatagram
///////////////////////////////////////////////////////////////////////////////

CTestDatagram::CTestDatagram(
   const DWORD id,
   const DataLength messageSize)
   :  m_id(id),
      m_messageSize(messageSize),
      m_pData(new BYTE[messageSize]),
      m_repeatCount(0),
      m_currentSendIteration(0),
      m_currentRecvIteration(0)
{
   memcpy(m_pData, &id, sizeof(id));

   for (auto i = sizeof(id); i < m_messageSize; ++i)
   {
      if (i % 2)
      {
         m_pData[i] = static_cast<BYTE>(i);
      }
      else
      {
         m_pData[i] = static_cast<BYTE>(id);
      }
   }
}

CTestDatagram::~CTestDatagram()
{
   delete [] m_pData;
}

// Implement ITestDatagram

void CTestDatagram::ExpectRepeats(
   const size_t numberOfRepeats)
{
   m_times.SetNumberOfTimes(numberOfRepeats);

   m_repeatCount = numberOfRepeats;
}

DWORD CTestDatagram::GetIdentifier() const
{
   return m_id;
}

const BYTE * const CTestDatagram::GetBytes()
{
   if (m_currentSendIteration >= m_repeatCount)
   {
      throw CException(
         _T("CTestDatagram::GetBytes()"),
         _T("m_currentSendIteration >= m_repeatCount: ") +
         ToString(m_currentSendIteration) + _T(" >= ") + ToString(m_repeatCount));
   }

   m_times.MarkAsSent(m_currentSendIteration++);

   return m_pData;
}

CTestDatagram::DataLength CTestDatagram::GetSize() const
{
   return m_messageSize;
}

static _tstring GetDataMismatch(
   const BYTE * const pData,
   const BYTE * const pExpectedData,
   const CTestDatagram::DataLength size,
   const CTestDatagram::DataLength contextLength,
   const CTestDatagram::DataLength maxMismatchLength)
{
   auto dataMatches = true;

   CTestDatagram::DataLength firstMismatch = 0;

   for (CTestDatagram::DataLength i = 0; i < size && dataMatches; ++i)
   {
      if (pData[i] != pExpectedData[i])
      {
         firstMismatch = i;
         dataMatches = false;
      }
   }

   if (!dataMatches)
   {
      auto lastMismatch = size;

      for (auto i = firstMismatch; i < size && !dataMatches; ++i)
      {
         if (pData[i] == pExpectedData[i])
         {
            lastMismatch = i;
            dataMatches = true;
         }
      }

      if (firstMismatch > contextLength)
      {
         firstMismatch -= contextLength;
      }
      else
      {
         firstMismatch = 0;
      }

      if (lastMismatch - firstMismatch > maxMismatchLength)
      {
         lastMismatch = firstMismatch + maxMismatchLength;
      }
      else
      {
         if (size - lastMismatch > contextLength)
         {
            lastMismatch += contextLength;
         }
         else
         {
            lastMismatch = size;
         }
      }

      const size_t dumpLength = lastMismatch - firstMismatch + 1;

      auto result = _T("Got: ") + DumpData(pData + firstMismatch, dumpLength);

      result += _T(" Expected: ") + DumpData(pExpectedData + firstMismatch, dumpLength);

      return result;
   }

   return _T("");
}

void CTestDatagram::ValidateResponse(
   const BYTE * const pData,
   const DataLength dataLength)
{
   if (m_currentRecvIteration > m_repeatCount)
   {
      throw CException(_T("CTestDatagram::ValidateResponse()"), _T("m_currentRecvIteration > m_repeatCount"));
   }

   m_times.MarkAsReceived(m_currentRecvIteration++);

   if (dataLength != m_messageSize)
   {
      throw CException(_T("CTestDatagram::ValidateResponse()"), _T("Response datagram is a different size: ") + ToString(dataLength) + _T(" expected: ") + ToString(m_messageSize));
   }

   if (0 != memcmp(pData, m_pData, dataLength))
   {
      const DataLength contextLength = 10;
      const DataLength maxMismatchLength = 20;

      const auto dataMismatch = GetDataMismatch(pData, m_pData, dataLength, contextLength, maxMismatchLength);

      Output(dataMismatch);

      Output(_T("     Got: ") + DumpData(pData, dataLength));
      Output(_T("Expected: ") + DumpData(m_pData, dataLength));

      throw CException(_T("CTestDatagram::ValidateResponse()"), _T("Data does not validate: ") + dataMismatch);
   }
}

const CRoundTripTimes &CTestDatagram::GetTimes() const
{
   return m_times;
}

///////////////////////////////////////////////////////////////////////////////
// End of file: TestDatagram.cpp
///////////////////////////////////////////////////////////////////////////////
