///////////////////////////////////////////////////////////////////////////////
// File: TestDatagrams.cpp
///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2009 JetByte Limited.
//
// This software is provided "as is" without a warranty of any kind. All
// express or implied conditions, representations and warranties, including
// any implied warranty of merchantability, fitness for a particular purpose
// or non-infringement, are hereby excluded. JetByte Limited and its licensors
// shall not be liable for any damages suffered by licensee as a result of
// using the software. In no event will JetByte Limited be liable for any
// lost revenue, profit or data, or for direct, indirect, special,
// consequential, incidental or punitive damages, however caused and regardless
// of the theory of liability, arising out of the use of or inability to use
// software, even if JetByte Limited has been advised of the possibility of
// such damages.
//
///////////////////////////////////////////////////////////////////////////////

#include "JetByteTools/Admin/Admin.h"

#include "TestDatagrams.h"
#include "TestDatagram.h"

#include "JetByteTools/CoreTools/FileUtils.h"
#include "JetByteTools/CoreTools/StringUtils.h"
#include "JetByteTools/CoreTools/Exception.h"

#include "JetByteTools/SocketTools/IDatagramSocket.h"

#pragma hdrstop

#include <iostream>

#include <algorithm>    // for min

///////////////////////////////////////////////////////////////////////////////
// Using directives
///////////////////////////////////////////////////////////////////////////////

using JetByteTools::Milliseconds;

using JetByteTools::Core::CException;
using JetByteTools::Core::ToString;
using JetByteTools::Core::_tstring;

using JetByteTools::Socket::IDatagramSocket;

using std::cout;
using std::endl;
using std::min;

///////////////////////////////////////////////////////////////////////////////
// Static helper functions
///////////////////////////////////////////////////////////////////////////////

static void DumpRTT(
   const CRoundTripTimes &times);

///////////////////////////////////////////////////////////////////////////////
// CTestDatagrams
///////////////////////////////////////////////////////////////////////////////

CTestDatagrams::CTestDatagrams(
   const DWORD repeatCount)
   :  m_repeatCount(repeatCount),
      m_succeeded(true),
      m_repliesPending(0)
{

}

CTestDatagrams::~CTestDatagrams()
{
   for (auto it = m_datagrams.begin(); it != m_datagrams.end(); ++it)
   {
      auto pDatagram = *it;

      delete pDatagram;
   }
}

DWORD CTestDatagrams::GetNumberOfRepliesPending() const
{
   return m_repliesPending;
}

HANDLE CTestDatagrams::GetWaitHandle() const
{
   return m_completeEvent.GetWaitHandle();
}

bool CTestDatagrams::HasFailed() const
{
   return !m_succeeded;
}

void CTestDatagrams::Fail()
{
   m_succeeded = false;

   m_completeEvent.Set();
}

bool CTestDatagrams::DataflowSucceeded() const
{
   return m_succeeded && DataflowComplete();
}

bool CTestDatagrams::DataflowComplete() const
{
   return m_completeEvent.IsSignalled();
}

void CTestDatagrams::AddDatagram(
   ITestDatagram *pDatagram)
{
   const size_t totalReplies = m_repliesPending + m_repeatCount;

   if (totalReplies > std::numeric_limits<DWORD>::max())
   {
      throw CException(
         _T("CTestDatagrams::AddDatagram()"),
         _T("Too many datagrams repeating too many times!"));
   }

   m_datagrams.push_back(pDatagram);

   pDatagram->ExpectRepeats(m_repeatCount);

   m_repliesPending = static_cast<DWORD>(totalReplies);
}

ITestDatagram &CTestDatagrams::GetDatagram(
   const DWORD id) const
{
   if (id > m_datagrams.size())
   {
      throw CException(_T("CTestDatagrams::GetDatagram()"), _T("Index out of range"));
   }

   return *m_datagrams[id];
}

void CTestDatagrams::ValidateReceivedData(
   const BYTE *pData,
   const DataLength dataLength)
{
   if (dataLength < sizeof(DWORD))
   {
      throw CException(_T("CTestDatagrams::ValidateReceivedData()"), _T("Invalid datagram size, cannot be smaller than sizeof(DWORD)"));
   }

   const auto id = *reinterpret_cast<const DWORD *>(pData);

   auto &datagram = GetDatagram(id);

   datagram.ValidateResponse(pData, dataLength);

   if (0 == ::InterlockedDecrement(&m_repliesPending))
   {
      m_completeEvent.Set();
   }
}

size_t CTestDatagrams::GetNumSent() const
{
   size_t count = 0;

   for (auto it = m_datagrams.begin(), end = m_datagrams.end();
      it != end;
      ++it)
   {
      count += (*it)->GetTimes().GetNumSent();
   }

   return count;
}

size_t CTestDatagrams::GetNumReceived() const
{
   size_t count = 0;

   for (auto it = m_datagrams.begin(), end = m_datagrams.end();
      it != end;
      ++it)
   {
      count += (*it)->GetTimes().GetNumReceived();
   }

   return count;
}

DWORD CTestDatagrams::GetAverageRTT() const
{
   size_t averageRTT = 0;

   for (auto it = m_datagrams.begin(), end = m_datagrams.end();
      it != end;
      ++it)
   {
      averageRTT += (*it)->GetTimes().GetAverageRTT();
   }

   return static_cast<DWORD>(averageRTT / m_datagrams.size());
}

DWORD CTestDatagrams::GetMaxRTT() const
{
   DWORD maxRTT = 0;

   for (auto it = m_datagrams.begin(), end = m_datagrams.end();
      it != end;
      ++it)
   {
      const auto rtt = (*it)->GetTimes().GetMaxRTT();

      if (rtt > maxRTT)
      {
         maxRTT = rtt;
      }
   }

   return maxRTT;
}

DWORD CTestDatagrams::GetMinRTT() const
{
   DWORD minRTT = MAXDWORD;

   for (auto it = m_datagrams.begin(), end = m_datagrams.end();
      it != end;
      ++it)
   {
      const auto rtt = (*it)->GetTimes().GetMinRTT();

      if (rtt < minRTT)
      {
         minRTT = rtt;
      }
   }

   return minRTT;
}

void CTestDatagrams::DumpRTT() const
{
   for (auto it = m_datagrams.begin(), end = m_datagrams.end();
      it != end;
      ++it)
   {
      auto pDatagram = *it;

      ::DumpRTT(pDatagram->GetTimes());
   }
}

///////////////////////////////////////////////////////////////////////////////
// Static helper functions
///////////////////////////////////////////////////////////////////////////////

static void DumpRTT(
   const CRoundTripTimes &times)
{
   for (auto it = times.Begin(), end = times.End();
      it != end;
      ++it)
   {
      cout << it->GetStatus() << ", " << it->GetSentTime() << ", " << it->GetReceivedTime() << ", " << it->GetRoundTripTime() << endl;
   }
}

///////////////////////////////////////////////////////////////////////////////
// End of file: TestDatagrams.cpp
///////////////////////////////////////////////////////////////////////////////
