///////////////////////////////////////////////////////////////////////////////
// File: TestMain.cpp
///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2006 JetByte Limited.
//
// This software is provided "as is" without a warranty of any kind. All
// express or implied conditions, representations and warranties, including
// any implied warranty of merchantability, fitness for a particular purpose
// or non-infringement, are hereby excluded. JetByte Limited and its licensors
// shall not be liable for any damages suffered by licensee as a result of
// using the software. In no event will JetByte Limited be liable for any
// lost revenue, profit or data, or for direct, indirect, special,
// consequential, incidental or punitive damages, however caused and regardless
// of the theory of liability, arising out of the use of or inability to use
// software, even if JetByte Limited has been advised of the possibility of
// such damages.
//
///////////////////////////////////////////////////////////////////////////////

#include "JetByteTools/Admin/Admin.h"

#include "JetByteTools/SocketTools/WinsockWrapper.h"

#include "JetByteTools/CoreTools/Exception.h"
#include "JetByteTools/CoreTools/FileUtils.h"
#include "JetByteTools/CoreTools/StringUtils.h"
#include "JetByteTools/CoreTools/DebugTrace.h"
#include "JetByteTools/CoreTools/StringConverter.h"
#include "JetByteTools/Win32Tools/GlobalName.h"
#include "JetByteTools/Win32Tools/ManualResetEvent.h"

#include "DatagramSocketConnectionManager.h"

#pragma hdrstop

#include "CommandLine.h"
#include "TestDatagrams.h"
#include "TestDatagram.h"

#include "JetByteTools/CoreTools/PerformanceCounter.h"
#include "JetByteTools/Win32Tools/SecurityDescriptorAllowAll.h"
#include "JetByteTools/Win32Tools/SecurityAttributes.h"
#include "JetByteTools/CoreTools/GlobalErrorHandler.h"

#include "JetByteTools/IOTools/IOPool.h"
#include "JetByteTools/IOTools/BufferAllocator.h"

#include "JetByteTools/SocketTools/FullAddress.h"
#include "JetByteTools/SocketTools/AddressTypeIPv4.h"
#include "JetByteTools/SocketTools/AddressIPv4.h"
#include "JetByteTools/SocketTools/DatagramSocketAllocator.h"

#include <iostream>

#include <algorithm>    // for min

///////////////////////////////////////////////////////////////////////////////
// Using directives
///////////////////////////////////////////////////////////////////////////////

using JetByteTools::Milliseconds;

using JetByteTools::Core::CException;
using JetByteTools::Core::ToString;
using JetByteTools::Core::_tstring;
using JetByteTools::Core::CStringConverter;
using JetByteTools::Win32::CGlobalName;
using JetByteTools::Win32::CManualResetEvent;
using JetByteTools::Core::CPerformanceCounter;
using JetByteTools::Win32::CSecurityDescriptorAllowAll;
using JetByteTools::Win32::CSecurityAttributes;
using JetByteTools::Core::CGlobalErrorHandler;

using JetByteTools::Socket::CFullAddress;
using JetByteTools::Socket::CAddressTypeIPv4;
using JetByteTools::Socket::CAddressType;
using JetByteTools::Socket::CAddressIPv4;
using JetByteTools::Socket::CDatagramSocketAllocator;

using JetByteTools::IO::CIOPool;
using JetByteTools::IO::CBufferAllocator;

using std::cout;
using std::endl;
using std::min;

///////////////////////////////////////////////////////////////////////////////
// Program entry point
///////////////////////////////////////////////////////////////////////////////

int main(int argc, const char **argv)
{
   auto result = 2;

   CGlobalErrorHandler errorHandler;

   CCommandLine commandLine;

   if (commandLine.Parse(argc, argv))
   {
      result = 1;

      try
      {
         const auto numRepeats = commandLine.Repeats();

         const auto numPackets = commandLine.Packets();

         const auto packetSize = commandLine.PacketSize();

         const auto delayAfter = commandLine.PacketBatchSize();

         const auto delayFor = commandLine.PacketBatchPause();

         const auto maxSocketsInPool = numPackets;

         const auto maxBuffersInPool = static_cast<CCommandLine::Count>(numPackets * 2);

         const auto bufferSize = packetSize;

         CDatagramSocketAllocator socketAllocator(
            maxSocketsInPool,
            commandLine.MaxBuffersPooledPerSocket(),
            commandLine.MaxConsecutiveOperationsOnOneSocket());

         CBufferAllocator bufferAllocator(
            bufferSize,
            maxBuffersInPool);

         CIOPool ioPool(
            0);                           // number of threads (0 = 2 x processors)

         ioPool.Start();

         cout << "Building " << numPackets << " packets" << endl;

         CTestDatagrams datagrams(commandLine.Repeats());

         for (DWORD i = 0; i < numPackets; ++i)
         {
            datagrams.AddDatagram(new CTestDatagram(i, packetSize));
         }

         CDatagramSocketConnectionManager connector(
            ioPool,
            socketAllocator,
            bufferAllocator,
            datagrams,
            commandLine.ReadTimeout());

         const auto addressPreference = CAddressTypeIPv4::instance;

         const CFullAddress address(
            commandLine.Server(),
            commandLine.Port(),
            addressPreference);

         if (commandLine.PreAllocate())
         {
            socketAllocator.EnlargePool(maxSocketsInPool);

            bufferAllocator.EnlargePool(maxBuffersInPool);
         }

         try
         {
            cout << "Sending " << numPackets << " packets ";

            if (numRepeats > 1)
            {
               cout << numRepeats << " times";
            }

            cout << endl;

            CPerformanceCounter counter(CPerformanceCounter::LockThreadToOneCPU);

            for (auto y = 0; y < numRepeats; ++y)
            {
               for (auto i = 0; i < numPackets && !datagrams.HasFailed(); ++i)
               {
                  auto &datagram = datagrams.GetDatagram(i);

                  connector.SendTo(address, datagram.GetBytes(), datagram.GetSize());

                  if (delayAfter && ((i + 1) % delayAfter == 0))
                  {
                     cout << i + 1 << " packets sent" << endl;

                     if (delayFor)
                     {
                        Sleep(delayFor);
                     }
                  }
               }
            }

            if (datagrams.HasFailed())
            {
               cout << "Packet echo timeout during sending..." << endl;
            }
            else
            {
               cout << "All packet echoes in progress..." << endl;
            }

            CSecurityDescriptorAllowAll sd;

            CSecurityAttributes sa(sd);

            CManualResetEvent shutdownEvent(CGlobalName(_T("JetByteToolsServerShutdown")), &sa);

            cout << "Waiting for data flow to complete" << endl;

            HANDLE handlesToWaitFor[2];

            handlesToWaitFor[0] = shutdownEvent.GetWaitHandle();
            handlesToWaitFor[1] = datagrams.GetWaitHandle();

            const auto checkEvery = commandLine.DisplayStatusEvery();

            Milliseconds elapsedTime = 0;

            auto done = false;

            const auto dataFlowTimeoutMillis = commandLine.DataFlowTimeout();

            while (!done)
            {
               const auto timeout = (dataFlowTimeoutMillis == INFINITE ? checkEvery : min(checkEvery, dataFlowTimeoutMillis - elapsedTime));

               elapsedTime += timeout;

               auto waitResult = ::WaitForMultipleObjects(2, handlesToWaitFor, false, timeout);

               if (waitResult == WAIT_TIMEOUT)
               {
                  if (elapsedTime >= dataFlowTimeoutMillis)
                  {
                     cout << "Waiting for data flow to complete timed out after " << CStringConverter::TtoA(counter.GetElapsedTimeAsString()) << endl;

                     result = 1;
                     done = true;
                  }
                  else
                  {
                     cout << "Waiting for " << datagrams.GetNumberOfRepliesPending() << " packets to be echoed" << endl;

                     connector.DumpErrors();
                  }
               }
               else if (waitResult == WAIT_OBJECT_0)
               {
                  result = 1;
                  done = true;
               }
               else if (waitResult == WAIT_OBJECT_0 + 1)
               {
                  const auto elapsedTime = counter.GetElapsedTimeAsString();

                  cout << "Data flow complete in " << CStringConverter::TtoA(elapsedTime) << endl;

                  cout << "   Packets sent: " << datagrams.GetNumSent() << endl;
                  cout << "Echoes received: " << datagrams.GetNumReceived() << endl;
                  cout << "    Average RTT: " << datagrams.GetAverageRTT() << "ms" << endl;
                  cout << "        Max RTT: " << datagrams.GetMaxRTT() << "ms" << endl;
                  cout << "        Min RTT: " << datagrams.GetMinRTT() << "ms" << endl;

                  result = 0;
                  done = true;

                  result = datagrams.DataflowSucceeded() ? 0 : 1;

                  if (!connector.DumpErrors())
                  {
                     result = 1;
                  }

                  if (commandLine.DumpRTT())
                  {
                     datagrams.DumpRTT();
                  }
               }
               else
               {
                  throw CException(_T("main()"), _T("Unexpected result from WaitForMultipleObjects"));
               }
            }
         }
         catch(const CException &e)
         {
            cout << "Exception: " << e.GetDetailsA() << endl;

            result = 1;
         }
         catch(...)
         {
            cout << "Unexpected exception" << endl;

            result = 1;
         }

         cout << "Wait for connector shutdown..." << endl;

         if (!connector.WaitForShutdownToComplete(10000))
         {
            cout << "Connector shutdown timedout!" << endl;

            result = 1;

            connector.ForceShutdown();
         }

         cout << "Wait for IO pool shutdown..." << endl;

         ioPool.WaitForShutdownToComplete();

         cout << "Waiting for socket allocator to flush..." << endl;

         socketAllocator.ReleaseSockets();

         cout << "Waiting for buffer allocator to flush..." << endl;

         bufferAllocator.Flush();

         if (!connector.DumpErrors())
         {
            result = 1;
         }

         cout << "All done..." << endl;
      }
      catch(const CException &e)
      {
         cout << "Exception: " << CStringConverter::TtoA(e.GetDetails()) << endl;

         result = 1;
      }
      catch(const char *pMsg)
      {
         cout << pMsg << endl;

         result = 1;
      }
      catch(...)
      {
         cout << "Unexpected exception" << endl;

         result = 1;
      }
      cout << "Test " << (result == 0 ? "Passed" : "Failed") << endl;
   }

   return result;
}

///////////////////////////////////////////////////////////////////////////////
// End of file: ServerMain.cpp
///////////////////////////////////////////////////////////////////////////////
